import {
  getChunks,
  isChunkExpired,
  setLocalFileInfo,
  removeLocalFileInfo,
  getLocalFileInfo,
  isContainFileMimeType,
  sum,
  request,
  computeMd5,
  sendTracking
} from "./utils";
import { updateProgress } from "./progressbeforeupload";
import { Pool } from "./pool";
import env from './env'

// 文件分块大小
// let BLOCK_SIZE = 2m * 1024 * 1024;
// 限制文件上传类型
const ERR_CODE_FILE_TYPE = 10001;
// 视频
const ERR_CODE_FILE_VIDEO_SIZE = 10002;
// 图片
const ERR_CODE_FILE_IMAGE_SIZE = 10003;
// 上传失败
const ERR_CODE_SEND_TYPE = 10004;
// 图片格式
const imageType = ["image/png", "image/jpeg", "image/jpg"];
// 视频格式
const videoType = ["video/mp4"];
// 文本格式
const txtType = [
  "text/plain",
  "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
  "application/vnd.ms-excel",
  "text/csv",
  "text/xml",
  "application/xml"
];

//
const REQUEST_SOURCE_EXCLUDES = ["fanstop", "kabp", "superfansfile"];

//upload_id过期
const ERR_CODE_INVALID_UPLOADID = 10005;
//缺少customerid 或 source 参数

const ERR_CODE_QUYER = 10006;
const ERR_CODE_QUYER_MSG = "缺少参数";

const host = env.get('host');
const intranet_host = env.get('intranet_host');
let MAX = 512;

let MAX_IMG = 5;

const imageUploadApi = "/api/v1/image/upload";
const imageUploadApiWep = "/api/v1/image/uploadWebp";
export class UploadManager {
  constructor(options, handlers, statisticsLogger) {
    this.config = Object.assign(
      {
        retryCount: 3,
        checkByMD5: false,
        forceDirect: false,
        concurrentRequestLimit: 3
      },
      options.config
    );
    this.extension = Object.assign(
      {
        fname: "",
        params: {},
        mimeType: null,
        downLoadMimeType: null,
        blockSize: 10 * 1024 * 1024
      },
      options.extension
    );
    this.statisticsLogger = statisticsLogger;
    this.progress = null;
    this.xhrList = [];
    this.xhrHandler = xhr => this.xhrList.push(xhr);
    this.aborted = false;
    this.file = options.file;
    this.key = options.key;
    this.onData = () => {};
    this.onError = () => {};
    this.onComplete = () => {};
    this.retryCount = 0;
    Object.assign(this, handlers);
  }

  putFile() {
    this.aborted = false;
    if (!isNaN(this.extension.maxSize)) {
      MAX = this.extension.maxSize;
    }
    if (!isNaN(this.extension.maxSizeImg)) {
      MAX_IMG = this.extension.maxSizeImg;
    }
    // 文件名称
    if (!this.extension.fname) {
      this.extension.fname = this.file.name;
    }
    // 文件类型
    if (this.extension.mimeType && this.extension.mimeType.length) {
      if (!isContainFileMimeType(this.file.type, this.extension.mimeType)) {
        this.onError({
          code: ERR_CODE_FILE_TYPE,
          msg: "文件类型错误"
        });
        return;
      }
    }
    // 文件大小 图片不超过5m 视频不超过500m
    if (/image/.test(this.file.type)) {
      if (this.file.size > MAX_IMG * 1024 * 1024) {
        this.onError({
          code: ERR_CODE_FILE_IMAGE_SIZE,
          msg: `图片大小不能超过${MAX_IMG}M`
        });
      }
    } else if (/video/.test(this.file.type)) {
      if (this.file.size > MAX * 1024 * 1024) {
        this.onError({
          code: ERR_CODE_FILE_VIDEO_SIZE,
          msg: `视频大小不能超过 ${
            MAX < 1024 ? `${MAX}M` : `${(MAX / 1024).toFixed(1)}G`
          }`
        });
        return;
      }
    } else if (/text/.test(this.file.type)) {
      if (this.file.size > 1024 * 1024 * 1024) {
        this.onError({
          code: ERR_CODE_FILE_VIDEO_SIZE,
          msg: "txt文件不能超过1024M"
        });
        return;
      }
    } else {
      if (this.file.size > 1024 * 1024 * 1024) {
        this.onError({
          code: ERR_CODE_FILE_VIDEO_SIZE,
          msg: "文件大小不能超过1024M"
        });
        return;
      }
    }
    sendTracking({
      eventid: 1099562, // 申请的事件id
      eventtype: "upload-file-before"
    });
    let startTime = new Date().getTime();
    let upload = new Promise((resolve, reject) => {
      this.uploadAt = new Date().getTime();
      updateProgress("upload");
      if (
        (this.extension.uploadMode === "webp" ||
          this.config.forceDirect ||
          /image/.test(this.file.type)) &&
        this.extension.uploadMode !== "file"
      ) {
        resolve(this.directUpload());
      } else {
        resolve(this.resumeUpload());
      }
    });
    upload.then(
      res => {
        let duration = new Date().getTime() - startTime;
        sendTracking({
          eventid: 1099562, // 申请的事件id
          eventtype: "upload-file-success",
          eventvalue: duration
        });
        this.onComplete(res);
      },
      err => {
        sendTracking({
          eventid: 1099562, // 申请的事件id
          eventtype: "upload-chunk-err", // 事件类型 ，非必须
          eventvalue: JSON.stringify(err)
        });
        this.clear();
        if (err.isRequestError && !this.config.disableStatisticsReport) {
          let reqId = this.aborted ? "" : err.reqId;
          let code = this.aborted ? -2 : err.code;
        }
        let needRetry = err.isRequestError && err.code === 0 && !this.aborted;
        let notReachRetryCount = ++this.retryCount <= this.config.retryCount;
        if (needRetry && notReachRetryCount) {
          this.putFile();
          return;
        }
        this.onError(err);
      }
    );
    return upload;
  }

  clear() {
    this.xhrList.forEach(xhr => xhr.abort());
    this.xhrList = [];
  }
  stop() {
    this.clear();
    this.aborted = true;
  }
  // 直传
  directUpload() {
    let formData = new FormData();
    formData.append("file", this.file);
    if (/image/.test(this.file.type)) {
      formData.append("picName", this.key);
    }
    this.extension.params.request_source &&
      formData.append("request_source", this.extension.params.request_source);
    formData.append("request_time", new Date().getTime());
    formData.append("customer_id", this.extension.params.customerId);
    // let domin = this.extension.params.isIntranet ? intranet_host : host;
    let domain = this.extension.requestDomain
      ? this.extension.requestDomain
      : this.extension.params.isIntranet
      ? intranet_host
      : host;
    if (this.extension.uploadMode === "webp") {
      let requestPath = imageUploadApiWep;
      return computeMd5(this.file).then(computedMd5 => {
        let md5 = this.extension.params.md5 || computedMd5;
        this.uploadUrl = `${domain}${requestPath}?customer_id=${this.extension.params.customerId}&request_source=${this.extension.params.request_source}&key=${this.extension.params.key}&pic_name=${this.extension.params.pic_name}&md5=${md5}`;
        if (this.extension.params.quality) {
          this.uploadUrl += `&quality=${this.extension.params.quality}`;
        }
        if (this.extension.params.lossLess) {
          this.uploadUrl += `&lossLess=${this.extension.params.lossLess}`;
        }
        return request(this.uploadUrl, {
          method: "POST",
          body: formData,
          withCredentials: !this.extension.params.isIntranet,
          onProgress: data => {
            this.updateDirectProgress(data.loaded, data.total);
          },
          onCreate: this.xhrHandler
        }).then(res => {
          this.finishDirectProgress();
          return res;
        });
      });
    } else {
      let requestPath = this.extension.requestPath || imageUploadApi;
      this.uploadUrl = `${domain}${requestPath}?customer_id=${this.extension.params.customerId}&request_source=${this.extension.params.request_source}`;
      return request(this.uploadUrl, {
        method: "POST",
        body: formData,
        withCredentials: !this.extension.params.isIntranet,
        onProgress: data => {
          this.updateDirectProgress(data.loaded, data.total);
        },
        onCreate: this.xhrHandler
      }).then(res => {
        this.finishDirectProgress();
        return res;
      });
    }
  }
  // 分片上传
  resumeUpload() {
    this.loaded = {
      mkFileProgress: 0,
      chunks: null
    };
    this.ctxList = [];
    this.localInfo = getLocalFileInfo(this.file);
    this.chunks = getChunks(this.file, this.extension.params.blockSize);
    this.initChunksProgress();
    let pool = new Pool(
      chunkInfo => this.uploadChunk(chunkInfo),
      this.config.concurrentRequestLimit
    );
    let uploadChunks = this.chunks.map((chunk, index) => {
      return pool.enqueue({
        chunk,
        index
      });
    });
    // 所有任务上传完成
    let result = Promise.all(uploadChunks).then(res => {
      return this.mkFileReq(res);
    });
    let __this = this;
    // 所有任务上传完成
    result.then(
      res => {
        window.localStorage.removeItem(this.extension.params.md5);
        removeLocalFileInfo(this.file);
      },
      err => {
        // removeLocalFileInfo(this.file);

        return;
      }
    );
    return result;
  }

  uploadChunk(chunkInfo) {
    let { index, chunk } = chunkInfo;
    let info = this.localInfo[index];
    let requestUrl;
    if (
      !this.extension.params.request_source ||
      !this.extension.params.customerId
    ) {
      return Promise.reject({
        code: ERR_CODE_QUYER,
        msg: ERR_CODE_QUYER_MSG
      });
    }
    // let domin = this.extension.params.isIntranet ? intranet_host : host;
    let domain = this.extension.requestDomain
      ? this.extension.requestDomain
      : this.extension.params.isIntranet
      ? intranet_host
      : host;
    if (/video/.test(this.file.type) && this.extension.uploadMode !== "file") {
      let requestPath =
        this.extension.requestPath || "/api/v1/video/media/upload";
      requestUrl = `${domain}${requestPath}?customer_id=${this.extension.params.customerId}&request_source=${this.extension.params.request_source}`;
    } else {
      let fileRequestPath = this.extension.requestPath || "/api/v1/file/upload";
      requestUrl = `${domain}${fileRequestPath}?customer_id=${this.extension.params.customerId}&request_source=${this.extension.params.request_source}`;
      if (this.extension.params.upload_type) {
        requestUrl += `&upload_type=${this.extension.params.upload_type}`;
      }
    }

    // 是否过期
    let savedReusable = info && !isChunkExpired(info.time);
    let shouldCheckMD5 = this.config.checkByMD5;
    let reuseSaved = () => {
      this.updateChunkProgress(chunk.size, index);
      this.ctxList[index] = {
        ctx: info.ctx,
        size: info.size,
        time: info.time,
        md5: info.md5
      };
      return Promise.resolve("1");
    };

    if (savedReusable && !shouldCheckMD5) {
      return reuseSaved();
    }
    return computeMd5(chunk).then(md5 => {
      // 每个文件分块做md5 跟本地存储的文件块md5比对
      if (savedReusable && md5 === info.md5) {
        return reuseSaved();
      }
      let onProgress = data => {
        this.updateChunkProgress(data.loaded, index);
      };
      let onCreate = this.xhrHandler;
      let method = "POST";
      let __this = this;
      let fileSlice = new File([chunk], md5, {
        type: this.file.type
      });

      let formData = new FormData();
      if (
        !/video/.test(this.file.type) ||
        this.extension.uploadMode === "file"
      ) {
        //文件上传参数
        formData.append("name", __this.key);
        formData.append("file_length", __this.file.size);
        formData.append("check", __this.extension.params.md5.split("_")[0]);
        formData.append("file", fileSlice);
        formData.append("slice_check", md5);
        formData.append("part_number", index);
        formData.append("length", chunk.size);
        formData.append("upload_id", __this.extension.params.uploadId);
        formData.append("mimetype", __this.extension.downLoadMimeType);
        if (
          !REQUEST_SOURCE_EXCLUDES.includes(
            __this.extension.params.request_source
          )
        ) {
          formData.append(
            "request_source",
            __this.extension.params.request_source
          );
          formData.append("customer_id", __this.extension.params.customerId);
        }
        __this.extension.params.file_url &&
          formData.append("file_url", __this.extension.params.file_url);
      } else {
        //视频上传参数
        formData.append("file", fileSlice);
        formData.append("length", __this.extension.params.blockSize);
        formData.append("check", __this.extension.params.md5);
        formData.append("upload_id", __this.extension.params.uploadId);
        formData.append("part_number", index);
        formData.append("slice_check", md5);
        formData.append("file_length", chunk.size);
        formData.append(
          "request_source",
          __this.extension.params.request_source
        );
        formData.append("request_time", new Date().getTime());
        formData.append("customer_id", __this.extension.params.customerId);
      }
      sendTracking({
        eventid: 1099562, // 申请的事件id
        eventtype: "before-upload-chunk" // 事件类型 ，非必须
      });
      let startTime = new Date().getTime();
      return request(requestUrl, {
        method,
        body: formData,
        withCredentials: !__this.extension.params.isIntranet,
        onProgress,
        onCreate
      })
        .then(response => {
          onProgress({
            loaded: chunk.size
          });

          this.ctxList[index] = {
            time: new Date().getTime(),
            ctx: response.data.ctx,
            size: chunk.size,
            md5: md5
          };
          let eventvalue = {
            duration: new Date().getTime() - startTime,
            part_number: index,
            file_length: chunk.size,
            customer_id: __this.extension.params.customerId,
            upload_id: __this.extension.params.uploadId,
            request_source: __this.extension.params.request_source,
            md5:
              __this.extension.params.md5.split("_")[0] ||
              __this.extension.params.md5,
            ...response.data
          };
          sendTracking({
            eventid: 1099562, // 申请的事件id
            eventtype: "upload-chunk", // 事件类型 ，非必须
            eventvalue: JSON.stringify(eventvalue)
          });
          if (response.data && response.data.data) {
            setLocalFileInfo(this.file, this.ctxList);
          }
          return response;
        })
        .catch(err => {
          let eventvalue = {
            part_number: index,
            file_length: chunk.size,
            customer_id: __this.extension.params.customerId,
            upload_id: __this.extension.params.uploadId,
            request_source: __this.extension.params.request_source,
            sliceMd5: md5,
            md5:
              __this.extension.params.md5.split("_")[0] ||
              __this.extension.params.md5,
            error: err
          };
          sendTracking({
            eventid: 1099562, // 申请的事件id
            eventtype: "upload-chunk-request-err", // 事件类型 ，非必须
            eventvalue: JSON.stringify(eventvalue)
          });
        });
    });
  }

  mkFileReq(data) {
    let fid;
    if (!/video/.test(this.file.type)) {
      data.map(item => {
        if (item.downLoadUrl && item.downLoadUrl !== "") {
          fid = item.downLoadUrl;
        }
      });
    } else {
      data.map(item => {
        if (item.fid) {
          fid = item.fid;
        }
      });
    }
    return new Promise((resolve, reject) => {
      this.updateMkFileProgress(1);
      resolve(fid);
    });
  }

  updateDirectProgress(loaded, total) {
    // 当请求未完成时可能进度会达到100，所以total + 1来防止这种情况出现
    this.progress = {
      total: this.getProgressInfoItem(loaded, total + 1)
    };
    this.onData(this.progress);
  }

  finishDirectProgress() {
    // 在某些浏览器环境下，xhr 的 progress 事件无法被触发，progress 为 null，
    if (!this.progress) {
      this.progress = {
        total: this.getProgressInfoItem(this.file.size, this.file.size)
      };
      this.onData(this.progress);
      return;
    }

    let total = this.progress.total;
    this.progress = {
      total: this.getProgressInfoItem(total.loaded + 1, total.size)
    };
    this.onData(this.progress);
  }

  initChunksProgress() {
    this.loaded.chunks = this.chunks.map(_ => 0);
    this.notifyResumeProgress();
  }

  updateChunkProgress(loaded, index) {
    this.loaded.chunks[index] = loaded;
    this.notifyResumeProgress();
  }

  updateMkFileProgress(progress) {
    this.loaded.mkFileProgress = progress;
    this.notifyResumeProgress();
  }

  notifyResumeProgress() {
    this.progress = {
      total: this.getProgressInfoItem(
        sum(this.loaded.chunks) + this.loaded.mkFileProgress,
        this.file.size + 1
      ),
      chunks: this.chunks.map((chunk, index) => {
        return this.getProgressInfoItem(this.loaded.chunks[index], chunk.size);
      })
    };
    this.onData(this.progress);
  }

  getProgressInfoItem(loaded, size) {
    return {
      loaded: loaded,
      size: size,
      percent: (loaded / size) * 100
    };
  }
}
